/*

	This program is free software; you can redistribute it and/or modify

	it under the terms of the GNU General Public License version 2 

	as published by the Free Software Foundation.



	This program is distributed in the hope that it will be useful,

	but WITHOUT ANY WARRANTY; without even the implied warranty of

	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.



	You should have received a copy of the GNU General Public License

	along with this program; if not, write to the Free Software

	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA





	Copyright (C) 2006  Thierry Berger-Perrin <tbptbp@gmail.com>

*/

//

// Transcription to the run time version.

//





#include "specifics.h"

#include "kdlib_writer.h"



// debug_dump_problematic_node

#include <stdlib.h>	//FIX: debug_dump_problematic_node

#include "math_vec.h"

#include "math_aabb.h"

#include "math_triangle.h"



namespace kdlib {

	//FIXME: horrible kludge.

	extern const kdlib::triangle_t *triangle_soup_base;

}



class allocator_t {

public:

	allocator_t(kdtree::node_t *p, kdtree::tri_id_t *q)

		: curr_node(p), items(q), curr_item(q)

	{}



	// allocate 2 childs at once

	FINLINE

	kdtree::node_t *allocate_childs() {

		kdtree::node_t * const p = curr_node;

		curr_node += 2;

		return p;

	}



	// leaf content

	kdtree::tri_id_t *allocate_items(const uint_t count) {

		kdtree::tri_id_t * const p = curr_item;

		curr_item += count;

		return p;

	}



	uint32_t get_offset(kdtree::tri_id_t * const p) const {

		return uint32_t(p-items);

	}



	// for lack of a better place/mechanism.

	//FIXME: ugly.

	uint32_t get_triangle_id(const kdlib::triangle_t * const p) {

		return uint32_t(p - kdlib::triangle_soup_base);

	}



private:

	kdtree::node_t		*curr_node;

	kdtree::tri_id_t	* const items;

	kdtree::tri_id_t	*curr_item;

};





static void do_write(allocator_t &allocator, const kdlib::node_t * const src, kdtree::node_t * const dst) {

	if (src == 0) {

		sys::log("do_write: null src.\n");

	}

	else if (src->is_leaf) {

		// leaf.

		kdtree::tri_id_t * const p = allocator.allocate_items(src->count);

		const uint32_t

			count	= src->count,

			offset	= allocator.get_offset(p);



		uint_t written = 0, pairs = 0;

		kdlib::sideness_t side;

		const kdlib::lbox_t *q = src->items.extract(side);

		while (q) {

			if (side == kdlib::side_left) {

				p[written] = q->triangle_id;	//allocator.get_triangle_id(q->get_clip_box()->tri);

				++written;

			}

			else

				++pairs;



			q = q->dim[0][side].next.extract(side);	// hmm.

		}



		if (written != count) {

			sys::log("do_write: leaf... %d items but only %d written.\n", count, written);

			BREAKPOINT();

		}



		if (written != pairs) {

			sys::log("do_write: leaf... %d items, %d written but only %d matching sides.\n", count, written, pairs);

			BREAKPOINT();

		}



		dst->leaf.offset_flag	= offset | kdtree::node_t::mask_leaf;

		dst->leaf.count			= count;

	}

	else {

		// inner node.

		kdtree::node_t

			*left = allocator.allocate_childs(),

			*right = left + 1;



		const uint32_t diff = uint32_t(left-dst);



		dst->inner.dim_offset_flag	= src->split_axis | diff << 3;

		dst->inner.split_coord		= src->split_coord;



		do_write(allocator, src->left,	left);

		do_write(allocator, src->right,	right);

	}

}





bool_t kdlib::writer_t::write(kdlib::node_t *src) {

	//allocator_t	allocator(dst, ids);

	allocator_t	allocator(dst+2, ids); // root node + 1 to re align (cache wise)

	do_write(allocator, src, dst);

	return true;

}



bool_t kdlib::writer_t::init(const kdlib::compiler_t &compiler) {

	num_nodes	= compiler.stats.num_nodes+1;	// +1 cuz we re-align after the root node.

	num_ids		= compiler.stats.num_ids;



	//src			= &compiler.root;



	const int

		mem_nodes = sizeof(kdtree::node_t) * num_nodes,

		mem_ids = sizeof(kdtree::tri_id_t) * num_ids;



	dst = (kdtree::node_t *) sys::mem::allocate(mem_nodes);

	ids = (kdtree::tri_id_t *) sys::mem::allocate(mem_ids);



	const double mega = 1/(1024.*1024.);

	sys::log("kdlib::writer_t::init: %s %.3fM for nodes, %.3fM for ids.\n",

		((dst != 0) & (ids != 0)) ? "allocated" : "failed to allocate",

		double(mem_nodes)*mega, double(mem_ids)*mega);

	if (!(dst && ids))

		return false;





	return true;

}







void debug_dump_problematic_node(const char *filename, const kdlib::node_t * const node, const kdlib::triangle_t * const soup_base) {



	if (node->is_leaf) {

		 FILE *file = fopen(filename, "wb");

		 if (!file) return;



		const uint_t count = node->count;



		sys::log("debug_dump_problematic_node: dumping %d triangles to '%s'.\n", count, filename);





		fprintf(file, "mtllib debug.mtl\n");

		uint_t vidx = 1;



		// voxel

		fprintf(file, "usemtl voxel\n");

		fprintf(file, "g voxel\n");

		//debug_dump_aabb(file, node->cell);

		const aabb_t &box(node->cell);

		for (int i=0; i < 3; ++i) {

			const int k = i, ku = (i+1)%3, kv = (i+2)%3;

			vec_t v0, v1, v2, v3;



			v0[k] = box.min[k];

			v1[k] = box.min[k];

			v2[k] = box.min[k];

			v3[k] = box.min[k];



			v0[ku] = box.min[ku];

			v1[ku] = box.max[ku];

			v2[ku] = box.max[ku];

			v3[ku] = box.min[ku];



			v0[kv] = box.min[kv];

			v1[kv] = box.min[kv];

			v2[kv] = box.max[kv];

			v3[kv] = box.max[kv];



			fprintf(file, "v %f %f %f\n", v0.x, v0.y, v0.z);

			fprintf(file, "v %f %f %f\n", v1.x, v1.y, v1.z);

			fprintf(file, "v %f %f %f\n", v2.x, v2.y, v2.z);

			fprintf(file, "v %f %f %f\n", v3.x, v3.y, v3.z);



			fprintf(file, "f %d// %d// %d//\n", vidx+0, vidx+1, vidx+2);

			fprintf(file, "f %d// %d// %d//\n", vidx+2, vidx+3, vidx+0);

			vidx += 4;



			v0[k] = box.max[k];

			v1[k] = box.max[k];

			v2[k] = box.max[k];

			v3[k] = box.max[k];



			fprintf(file, "v %f %f %f\n", v0.x, v0.y, v0.z);

			fprintf(file, "v %f %f %f\n", v1.x, v1.y, v1.z);

			fprintf(file, "v %f %f %f\n", v2.x, v2.y, v2.z);

			fprintf(file, "v %f %f %f\n", v3.x, v3.y, v3.z);



			fprintf(file, "f %d// %d// %d//\n", vidx+0, vidx+1, vidx+2);

			fprintf(file, "f %d// %d// %d//\n", vidx+2, vidx+3, vidx+0);

			vidx += 4;

		}





		fprintf(file, "usemtl soup\n");

		fprintf(file, "g soup\n");

		uint_t written = 0;

		kdlib::sideness_t side;

		const kdlib::lbox_t *q = node->items.extract(side);

		while (q) {

			if (side == kdlib::side_left) {

				const kdlib::triangle_t &t(soup_base[q->triangle_id]);

				for (int i=0; i<3; ++i) {

					const vec_t &v(t.verts[i]);

					fprintf(file, "v %f %f %f\n", v.x, v.y, v.z);

				}

				fprintf(file, "f %d// %d// %d//\n", vidx+0, vidx+1, vidx+2);

				vidx += 3;



				++written;

			}



			q = q->dim[0][side].next.extract(side);	// hmm.

		}



		fclose(file);

	}

}

