/*

	This program is free software; you can redistribute it and/or modify

	it under the terms of the GNU General Public License version 2 

	as published by the Free Software Foundation.



	This program is distributed in the hope that it will be useful,

	but WITHOUT ANY WARRANTY; without even the implied warranty of

	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.



	You should have received a copy of the GNU General Public License

	along with this program; if not, write to the Free Software

	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA





	Copyright (C) 2006  Thierry Berger-Perrin <tbptbp@gmail.com>

*/

#ifndef MATH_UTILITY_H

#define MATH_UTILITY_H



/*

	we'll poll cmath/math.h in from here.

*/

#define _USE_MATH_DEFINES

#include <cmath>



#if 0

//#if defined __MSVC__

	// stupid fix for stupid compilers.

	static FINLINE float minf(const float &a, const float &b) { 

		float r;

		_mm_store_ss(&r, _mm_min_ss(_mm_load_ss(&a),_mm_load_ss(&b)));

		return r;

	}

	static FINLINE float maxf(const float &a, const float &b) {

		float r;

		_mm_store_ss(&r, _mm_max_ss(_mm_load_ss(&a),_mm_load_ss(&b)));

		return r;

	}

	static FINLINE void clampf(const float minima, const float maxima, const float input, float &output) { 

		const __m128  min = _mm_load_ss(&minima), max = _mm_load_ss(&maxima), in =  _mm_load_ss(&input);

		_mm_store_ss(&output, _mm_max_ss(min, _mm_min_ss(in, max)));

	}

#else

	static FINLINE float minf(const float a, const float b) { return (b > a) ? a : b; }

	static FINLINE float maxf(const float a, const float b) { return (b < a) ? a : b; }

	static FINLINE float clampf(const float minima, const float maxima, const float value) { return minf(maxf(value, minima), maxima); }

	static FINLINE void clampf(const float minima, const float maxima, const float value, float &output) { output = minf(maxf(value, minima), maxima); }

#endif



/*

	finally found some undocumented popcount & bit manip intrinsics for icc

	in ia32intrin.h



		// Implement RDTSC and RDPMC

		extern __int64 _rdtsc(void);

		extern __int64 _rdpmc(int);



		// Implements BSWAP 

		extern int _bswap(int);



		// Implement BSR and BSF

		extern int _bit_scan_forward(int);

		extern int _bit_scan_reverse(int);



		// Implements a 32-bit popcount

		extern int _popcnt32(int);

*/



// returns the number of 1-bit of x

static FINLINE int32_t bit_pop_count32(const uint32_t x) {

	#if defined __GCC__ || defined __ICC_GCC__

		return __builtin_popcount(x);

	#elif defined __ICC__

		return _popcnt32(x);

	#else

		// pffff.

		uint32_t num = 0, val = x;

		#ifdef __ICC__

			// hah, shit, icc completly unroll it. hahah.

			#pragma nounroll

		#endif

		for (uint_t i=0; i<32; ++i) {

			num += (val & 1) ? 1 : 0;

			val >>= 1;

		}



		return num;

	#endif

}



// returns the index of the least significant 1-bit of x

static FINLINE int32_t bit_scan_forward32(const uint32_t x) {

	#if defined __GCC__ || defined __ICC_GCC__
		return __builtin_ffs(x) - 1;
	#elif defined __MSVC__
		int32_t idx;
		_BitScanForward((DWORD*)&idx, x);
		return idx;
	#elif defined __ICC__
		return _bit_scan_forward(x);
		/*
		// pfff.
		#pragma nounroll

		for (uint_t i=0; i<32; ++i)

			if (x & (1<<i)) return i;

		return 0;

		*/

	#endif
}



/*

	Silly stuff



*/



// returns an integer going through <0,amplitude-1,0> with a (2*amplitude)-1 frequency

template <int amplitude> static int compute_cyclic(const int val) {

	enum { amp = amplitude-1, bias = amp, freq = 2*amp };

	return abs(amp - ((val+bias) % freq));

}



// mapped back to [0, 1]

template <int amplitude> static float compute_cyclicf(const int val) {

	enum { amp = amplitude-1, bias = amp, freq = 2*amp };

	return float(compute_cyclic<amplitude>(val))/float(amp);

}

/*

static int compute_cyclic(const int bias, const int amplitude, const int base_freq = 0) {

	const int part = (frames>>base_freq) % (2*amplitude);

	return bias + abs(part-amplitude);

}

*/







// should really put that into the constant section, but then we don't have much use for it so...

#ifndef M_PI

	#define M_PI	3.14159265358979323846
#endif

#define DEGRAD(a)	((a) * (float(M_PI) / 180.f))
#define RADDEG(a)	((a) * (180.f/ float(M_PI)))






#endif

