/*

	This program is free software; you can redistribute it and/or modify

	it under the terms of the GNU General Public License version 2 

	as published by the Free Software Foundation.



	This program is distributed in the hope that it will be useful,

	but WITHOUT ANY WARRANTY; without even the implied warranty of

	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.



	You should have received a copy of the GNU General Public License

	along with this program; if not, write to the Free Software

	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA





	Copyright (C) 2006  Thierry Berger-Perrin <tbptbp@gmail.com>

*/

#ifndef RT_CAMERA_H

#define RT_CAMERA_H



#include "specifics.h"

#include "math_vec.h"



#include "rt_framebuffer.h"

namespace rt {

	//

	// screen plane sampler (for shooting ray through)

	//

	/*

	struct sampler_t { 

		vec_t top, dx, dy;

		// float res_x,res_y;



		void set(const camera_t &cam, const point_t &res);



		// (x,y) in screen coords -> mapped to the camera ray gen plane.

		vec_t map(const float x, const float y) const { return vec_t(top + dx*x + dy*y); }



		FINLINE void generate_packet_pattern(const float px, const float py, vec4_t &v) const {
			// { 0,   +dx
			//   +dy, +dx+dy }
			const __m128 
				px2(set1ps(px)), px2_one(addps(px2, rt::cst::section.one)), x(orps(andps(rt::cst::section.packet_pattern_mask_x, px2), andnps(rt::cst::section.packet_pattern_mask_x, px2_one))),
				py2(set1ps(py)), py2_one(addps(py2, rt::cst::section.one)), y(orps(andps(rt::cst::section.packet_pattern_mask_y, py2), andnps(rt::cst::section.packet_pattern_mask_y, py2_one)));

			v.x.v = addps(set1ps(top[0]), addps(mulps(set1ps(dx[0]), x), mulps(set1ps(dy[0]),y)));
			v.y.v = addps(set1ps(top[1]), addps(mulps(set1ps(dx[1]), x), mulps(set1ps(dy[1]),y)));
			v.z.v = addps(set1ps(top[2]), addps(mulps(set1ps(dx[2]), x), mulps(set1ps(dy[2]),y)));
		}


	};

	*/



	//

	// camera

	// should match OpenGL for convenience.

	//

	class camera_t {

		vec_t	eye, dir, up, right;

		float	fovx;			// stored as half fov

		int		world_up_index;



	public:

		bool_t open();

		void close() {}



		NOINLINE void look_at(const vec_t &target, const int up_idx = -1);



		void set_fovx(const float degree) { fovx = DEGRAD(degree)/2.f; }

		void set_eye(const vec_t &v) { eye = v; }

		

		FINLINE float get_fovx()			const { return RADDEG(fovx)*2.f; }

		FINLINE const vec_t &get_eye()		const { return eye; }

		FINLINE const vec_t &get_up()		const { return up; }	// stand up

		FINLINE const vec_t &get_dir()		const { return dir; }

		FINLINE const vec_t &get_right()	const { return right; }



		void set_world_up_index(int idx)	{ world_up_index = idx; }

		int get_world_up_index()			const { return world_up_index; }



		struct sampler_t { 

			vec_t top, dx,dy;

			vec_t map(const point_t &screen) const { return vec_t(top + dx*float(screen.x) + dy*float(screen.y)); }

		} sampler;

		void update(const framebuffer_t &fb);

	};

}





#endif
